#include "tilda-config.h"

#include "tilda.h"
#include "tilda-types.h"
#include "debug.h"
#include "translation.h"

#include <glib-object.h> /* for GType */
#include <stdlib.h> /* for exit() */

GKeyFile *config_defaults = NULL;
GKeyFile *config_userprefs = NULL;

/*
 * TODO:
 * 1) Configuration forward-migration, automatically, on startup.
 */

/* The main idea behind this configuration system is that each type of GObject
 * that needs to use the system implements its own lookup scheme. I tried to
 * do this as generically as possible, by using just the GObject property names,
 * and then looking up their type and using that to use the correct parsing
 * function.
 *
 * It is basically function overloading (for example, from C++) but done in C.
 * Rather complicated, if you ask me. Also a complete pain. But the result
 * sure is nice.
 */

static gchar *
tilda_config_lookup_version (GKeyFile *keyfile, GError **error)
{
	debug_enter  ();

	return g_key_file_get_string (keyfile, "Meta", "version", error);
}

static gboolean
tilda_config_parse_integer (GKeyFile    *keyfile,
							const gchar *group_name,
							const gchar *key,
							GValue      *value,
							GError     **user_error)
{
	debug_enter  ();
	debug_assert (G_IS_VALUE(value));

	GError *error = NULL;
	gint ret;

	/* Try to get the value */
	ret = g_key_file_get_integer (keyfile, group_name, key, &error);

	/* Check for error */
	if (error)
	{
		g_propagate_error (user_error, error);
		return FALSE;
	}

	/* We successfully parsed the int, so set it in the GValue */
	g_value_set_int (value, ret);

	return TRUE;
}

static gboolean
tilda_config_parse_boolean (GKeyFile    *keyfile,
							const gchar *group_name,
							const gchar *key,
							GValue      *value,
							GError     **user_error)
{
	debug_enter  ();
	debug_assert (G_IS_VALUE(value));

	GError *error = NULL;
	gboolean ret;

	/* Try to get the value */
	ret = g_key_file_get_boolean (keyfile, group_name, key, &error);

	/* Check for error */
	if (error)
	{
		g_propagate_error (user_error, error);
		return FALSE;
	}

	/* Success */
	g_value_set_boolean (value, ret);

	return TRUE;
}

static gboolean
tilda_config_parse_string (GKeyFile    *keyfile,
						   const gchar *group_name,
						   const gchar *key,
						   GValue      *value,
						   GError **user_error)
{
	debug_enter  ();
	debug_assert (G_IS_VALUE(value));

	GError *error = NULL;
	gchar *ret;

	/* Try to get the value */
	ret = g_key_file_get_string (keyfile, group_name, key, &error);

	/* Check for error */
	if (error)
	{
		g_propagate_error (user_error, error);
		return FALSE;
	}

	/* Success */
	g_value_set_string (value, ret);

	return TRUE;
}

#define key_is(STR) (g_ascii_strcasecmp(key,(STR)) == 0)

static gboolean
tilda_config_parse_enum (GKeyFile    *keyfile,
						 const gchar *group_name,
						 const gchar *key,
						 GValue      *value,
						 GError     **user_error)
{
	gchar *ret;
	GError *error = NULL;
	GEnumClass *enum_class;
	GEnumValue *enum_value;

	if (key_is("backspace-binding") || key_is("delete-binding"))
		enum_class = g_type_class_peek (vte_terminal_erase_binding_get_type());
	else if (key_is("dynamic-title"))
		enum_class = g_type_class_peek (tilda_dynamic_title_get_type());
	else if (key_is("exit-action"))
		enum_class = g_type_class_peek (tilda_child_exit_action_get_type());
	else if (key_is("scrollbar-position"))
		enum_class = g_type_class_peek (tilda_scrollbar_position_get_type());
	else if (key_is("animation-orientation") || key_is("tab-position"))
		enum_class = g_type_class_peek (gtk_position_type_get_type());
	else
	{
		g_critical ("FIXME: developer error -- unknown enum key used: %s\n", key);
		exit (1);
		return FALSE;
	}

	/* Get the value from the config as a string */
	ret = g_key_file_get_string (keyfile, group_name, key, &error);

	if (error)
	{
		g_propagate_error (user_error, error);
		return FALSE;
	}

	enum_value = g_enum_get_value_by_nick (enum_class, ret);

	if (!enum_value)
	{
		/* This is exactly the same error that is returned by g_key_file_get_integer()
		 * when it cannot correctly parse the value. */
		g_set_error (user_error,
					 G_KEY_FILE_ERROR,
					 G_KEY_FILE_ERROR_INVALID_VALUE,
					 _("Key file contains key '%s' in group '%s' which has value that cannot be interpreted."), key, group_name);
		return FALSE;
	}

	/* All was successful, let's set it */
	g_value_set_enum (value, enum_value->value);
	return TRUE;
}

gboolean
tilda_controller_set_property_from_config (TildaController *self, const gchar *property)
{
	debug_enter  ();
	debug_assert (TILDA_IS_CONTROLLER(self));

	GError *error = NULL;
	gboolean ret;

	GParamSpec *pspec;
	GValue *value = g_new0 (GValue, 1);
	gboolean (*parse_func) (GKeyFile *keyfile, const gchar *group_name, const gchar *key, GValue *value, GError **error);

	/* Get the pspec for this property */
	pspec = g_object_class_find_property (G_OBJECT_GET_CLASS(self), property);

	/* Make sure that this property exists */
	if (pspec == NULL)
	{
		g_critical ("FIXME: developer error -- unable to find property: %s\n", property);
		exit (1);
	}

	/* Initialize the GValue that is going to hold the returned value */
	g_value_init (value, pspec->value_type);

	/* Set the correct function to do the parsing */
	if      (g_type_is_a (pspec->value_type, G_TYPE_INT))
		parse_func = tilda_config_parse_integer;
	else if (g_type_is_a (pspec->value_type, G_TYPE_BOOLEAN))
		parse_func = tilda_config_parse_boolean;
	else if (g_type_is_a (pspec->value_type, G_TYPE_STRING))
		parse_func = tilda_config_parse_string;
	else if (g_type_is_a (pspec->value_type, G_TYPE_ENUM))
		parse_func = tilda_config_parse_enum;
	else
	{
		g_critical ("FIXME: developer error -- unknown property type: %s\n", g_type_name(pspec->value_type));
		exit(1);
	}

	/* Do the [Controller] lookup */
	ret = parse_func (config_userprefs, "Controller", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [Global] lookup */
	ret = parse_func (config_userprefs, "Global", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [controller-defaults] lookup */
	ret = parse_func (config_defaults, "controller-defaults", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);

		/* This is somewhat of a nasty hack, but I really do want to just set the property to
		 * NULL if there is no default, but ONLY for strings. */
		if (parse_func == tilda_config_parse_string)
		{
			g_value_set_string (value, NULL);
			goto success;
		}
	}
	else
		goto success;

//failure:
	g_critical (_("Unable to find a value for controller property: %s\n"), property);
	g_free (value);
	return FALSE;

success:
	g_object_set_property (G_OBJECT(self), property, value);
	g_value_unset (value);
	g_free (value);
	return TRUE;
}

gboolean
tilda_window_set_property_from_config (TildaWindow *self, const gchar *property)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	gchar *group_name;
	GError *error = NULL;
	gboolean ret;

	GParamSpec *pspec;
	GValue *value = g_new0(GValue, 1);
	gboolean (*parse_func) (GKeyFile *keyfile, const gchar *group_name, const gchar *key, GValue *value, GError **error);

	/* Get the pspec for this property */
	pspec = g_object_class_find_property (G_OBJECT_GET_CLASS(self), property);

	/* Make sure that this property exists */
	if (pspec == NULL)
	{
		g_critical ("FIXME: developer error -- unable to find property: %s\n", property);
		exit (1);
	}

	/* Initialize the GValue that is going to hold the returned value */
	g_value_init (value, pspec->value_type);

	/* Set the correct function to do the parsing */
	if      (g_type_is_a (pspec->value_type, G_TYPE_INT))
		parse_func = tilda_config_parse_integer;
	else if (g_type_is_a (pspec->value_type, G_TYPE_BOOLEAN))
		parse_func = tilda_config_parse_boolean;
	else if (g_type_is_a (pspec->value_type, G_TYPE_STRING))
		parse_func = tilda_config_parse_string;
	else if (g_type_is_a (pspec->value_type, G_TYPE_ENUM))
		parse_func = tilda_config_parse_enum;
	else
	{
		g_critical ("FIXME: developer error -- unknown property type: %s\n", g_type_name(pspec->value_type));
		exit(1);
	}

	/* Do the [Window] lookup */
	group_name = g_strdup_printf ("Window%d", self->number);
	ret = parse_func (config_userprefs, group_name, property, value, &error);
	g_free (group_name);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [Global] lookup */
	ret = parse_func (config_userprefs, "Global", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [window-defaults] lookup */
	ret = parse_func (config_defaults, "window-defaults", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);

		/* This is somewhat of a nasty hack, but I really do want to just set the property to
		 * NULL if there is no default, but ONLY for strings. */
		if (parse_func == tilda_config_parse_string)
		{
			g_value_set_string (value, NULL);
			goto success;
		}
	}
	else
		goto success;

//failure:
	g_critical (_("Unable to find a value for window property: %s\n"), property);
	g_free (value);
	return FALSE;

success:
	g_object_set_property (G_OBJECT(self), property, value);
	g_value_unset (value);
	g_free (value);
	return TRUE;
}

gboolean
tilda_terminal_set_property_from_config (TildaTerminal *self, const gchar *property)
{
	debug_enter  ();
	debug_assert (TILDA_IS_TERMINAL(self));

	TildaWindow *parent_window = TILDA_WINDOW(self->parent_window);
	gchar *group_name;
	GError *error = NULL;
	gboolean ret;

	GParamSpec *pspec;
	GValue *value = g_new0 (GValue, 1);
	gboolean (*parse_func) (GKeyFile *keyfile, const gchar *group_name, const gchar *key, GValue *value, GError **error);

	/* Get the pspec for this property */
	pspec = g_object_class_find_property (G_OBJECT_GET_CLASS(self), property);

	/* Make sure that this property exists */
	if (pspec == NULL)
	{
		g_critical ("FIXME: developer error -- unable to find property: %s\n", property);
		exit (1);
	}

	/* Initialize the GValue that is going to hold the returned value */
	g_value_init (value, pspec->value_type);

	/* Set the correct function to do the parsing */
	if      (g_type_is_a (pspec->value_type, G_TYPE_INT))
		parse_func = tilda_config_parse_integer;
	else if (g_type_is_a (pspec->value_type, G_TYPE_BOOLEAN))
		parse_func = tilda_config_parse_boolean;
	else if (g_type_is_a (pspec->value_type, G_TYPE_STRING))
		parse_func = tilda_config_parse_string;
	else if (g_type_is_a (pspec->value_type, G_TYPE_ENUM))
		parse_func = tilda_config_parse_enum;
	else
	{
		g_critical ("FIXME: developer error -- unknown property type: %s\n", g_type_name(pspec->value_type));
		exit(1);
	}

	/* Do the [Window/Terminal] lookup */
	group_name = g_strdup_printf ("Window%d/Terminal%d", parent_window->number, self->number);
	ret = parse_func (config_userprefs, group_name, property, value, &error);
	g_free (group_name);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [Window] lookup */
	group_name = g_strdup_printf ("Window%d", parent_window->number);
	ret = parse_func (config_userprefs, group_name, property, value, &error);
	g_free (group_name);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [Global] lookup */
	ret = parse_func (config_userprefs, "Global", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);
	}
	else
		goto success;

	/* Do the [terminal-defaults] lookup */
	ret = parse_func (config_defaults, "terminal-defaults", property, value, &error);

	if (error)
	{
		if (error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
			g_warning (error->message);

		g_clear_error (&error);

		/* This is somewhat of a nasty hack, but I really do want to just set the property to
		 * NULL if there is no default, but ONLY for strings. */
		if (parse_func == tilda_config_parse_string)
		{
			g_value_set_string (value, NULL);
			goto success;
		}
	}
	else
		goto success;

//failure:
	g_critical (_("Unable to find a value for terminal property: %s\n"), property);
	g_free (value);
	return FALSE;

success:
	g_object_set_property (G_OBJECT(self), property, value);
	g_value_unset (value);
	g_free (value);
	return TRUE;
}

gboolean
tilda_config_init (const gchar *userprefs_filename)
{
	gchar *defaults_filename;
	GError *error = NULL;

	/* Call g_type_class_ref() on all enum types that are being used */
	g_type_class_ref (vte_terminal_erase_binding_get_type());
	g_type_class_ref (tilda_dynamic_title_get_type());
	g_type_class_ref (tilda_child_exit_action_get_type());
	g_type_class_ref (tilda_scrollbar_position_get_type());
	g_type_class_ref (gtk_position_type_get_type());

	/* Create the defaults file's path */
	defaults_filename = g_build_filename ("share-tilda.conf", NULL); // FIXME: use /usr/share/tilda

	/* Create the keyfiles */
	config_defaults = g_key_file_new ();
	config_userprefs = g_key_file_new ();

	/* Check if the defaults exist, and load them */
	if (!g_file_test (defaults_filename, G_FILE_TEST_EXISTS))
	{
		g_critical (_("No configuration defaults file found. Tilda may not work.\n"));
	}
	else
	{
		if (!g_key_file_load_from_file (config_defaults, defaults_filename, G_KEY_FILE_NONE, &error))
		{
			g_critical (_("Error reading configuration defaults: %s\n"), error->message);
			g_clear_error (&error);
		}
	}

	/* Check if the user's config exists, and load it */
	if (!g_file_test (userprefs_filename, G_FILE_TEST_EXISTS))
	{
		g_warning (_("No user configuration file found, using defaults\n"));
	}
	else
	{
		if (!g_key_file_load_from_file (config_userprefs, userprefs_filename, G_KEY_FILE_NONE, &error))
		{
			g_warning (_("Error reading user configuration: %s\n"), error->message);
			g_clear_error (&error);
		}
	}

	/* This is just here for the future. Currently, it is acceptable to run without
	 * a configuration, though Tilda may not work very well. */
	return TRUE;
}

/**
 * Since we DO NOT allow writing to the configuration file from a Tilda
 * process (only from the wizard), this does nothing more than free the
 * defaults hashtable, and free the opened keyfile.
 */
gboolean
tilda_config_free ()
{
	/* Since we never write the config file from within Tilda,
	 * we can just close the files. */
	g_key_file_free (config_defaults);
	g_key_file_free (config_userprefs);

	config_defaults = NULL;
	config_userprefs = NULL;

	return TRUE;
}

/* vim: set ts=4 sts=4 sw=4 noet tw=112: */
